/*=========================================================================

  Program:   SciVis 2006 - Exercise 2 - Flow Topology Analysis: Framework
  Module:    vtkFlowTopo.cxx
  Author:    Filip Sadlo

     This software is distributed WITHOUT ANY WARRANTY; without even
     the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
     PURPOSE.  See the above copyright notice for more information.

=========================================================================*/
#include "vtkFlowTopo.h"

#include "vtkCell.h"
#include "vtkIdList.h"
#include "vtkInformation.h"
#include "vtkInformationVector.h"
#include "vtkObjectFactory.h"
#include "vtkPointData.h"
#include "vtkPolyData.h"
#include "vtkCellArray.h"
#include "vtkPolyData.h"
#include "vtkCellType.h"

#include "linalg.h"

vtkCxxRevisionMacro(vtkFlowTopo, "$Revision: 0.01 $");
vtkStandardNewMacro(vtkFlowTopo);

const bool skipBoundaryCells = false;

vtkFlowTopo::vtkFlowTopo()
{
  GlyphRadius = 20.0;

  this->SetInputArrayToProcess(0, 0, 0, vtkDataObject::FIELD_ASSOCIATION_POINTS,
                               vtkDataSetAttributes::VECTORS);
}

vtkFlowTopo::~vtkFlowTopo()
{
}

/*
void generateCross(double posX, double posY, double posZ,
		   double glyphRadius,
		   double color,
		   vtkPoints *outputPoints,
		   vtkCellArray *outputLines,
		   vtkFloatArray *outputColors,
		   int &numPoints)
{
  // cross of three orthogonal lines
  // -> each line is represented by a VTK cell that contains 2 points

  for (int line=0; line<3; line++) {

    outputLines->InsertNextCell(2);
    for (int v=0; v<2; v++) {

      double pnt[3] = { posX, posY, posZ };
      if (v == 0) pnt[line] -= glyphRadius / 2.0;
      else pnt[line] += glyphRadius / 2.0;
    
      // store point
      outputPoints->InsertNextPoint(pnt);

      // insert point ID into cell (connectivity info)
      outputLines->InsertCellPoint(numPoints++);

      // colors
      outputColors->InsertNextValue(color);
    }
  }
}
*/

void generateLine(double startX, double startY, double startZ,
		  double endX, double endY, double endZ,
		  double color,
		  vtkPoints *outputPoints,
		  vtkCellArray *outputLines,
		  vtkFloatArray *outputColors,
		  int &numPoints)
{
  // single line
  // -> each line is represented by a VTK cell that contains 2 points

  outputLines->InsertNextCell(2);

  // store points
  double pnt[3] = { startX, startY, startZ };
  outputPoints->InsertNextPoint(pnt);
  double pnt2[3] = { endX, endY, endZ };
  outputPoints->InsertNextPoint(pnt2);

  // insert point IDs into cell (connectivity info)
  outputLines->InsertCellPoint(numPoints++);
  outputLines->InsertCellPoint(numPoints++); 

  // colors
  outputColors->InsertNextValue(color);
  outputColors->InsertNextValue(color);
}

void generateQuad(double luX, double luY, double luZ,
		  double ruX, double ruY, double ruZ,
		  double llX, double llY, double llZ,
		  double rlX, double rlY, double rlZ,
		  double color,
		  bool transparent,
		  vtkPoints *outputPoints,
		  vtkCellArray *outputLines,
		  vtkCellArray *outputPolygons,
		  vtkFloatArray *outputColors,
		  int &numPoints)
{
  // single quad, either opaque (polygon) or transparent (lines)
  // lu: left upper vertex of quad, etc.

  // store point
  double lu[3] = { luX, luY, luZ };
  double ru[3] = { ruX, ruY, ruZ };
  double ll[3] = { llX, llY, llZ };
  double rl[3] = { rlX, rlY, rlZ }; 

  outputPoints->InsertNextPoint(lu);
  outputPoints->InsertNextPoint(ru);
  outputPoints->InsertNextPoint(ll);
  outputPoints->InsertNextPoint(rl);
  
  int startId = numPoints;

  // polygon
  if (!transparent) {
    outputPolygons->InsertNextCell(4);
    outputPolygons->InsertCellPoint(numPoints++);
    outputPolygons->InsertCellPoint(numPoints++);
    outputPolygons->InsertCellPoint(numPoints++);
    outputPolygons->InsertCellPoint(numPoints++);
  }
  else {
    numPoints += 4;
  }

  // lines
  if (1) {
    // outer edges
    outputLines->InsertNextCell(2);
    outputLines->InsertCellPoint(startId + 0);
    outputLines->InsertCellPoint(startId + 1);

    outputLines->InsertNextCell(2);
    outputLines->InsertCellPoint(startId + 1);
    outputLines->InsertCellPoint(startId + 2);

    outputLines->InsertNextCell(2);
    outputLines->InsertCellPoint(startId + 2);
    outputLines->InsertCellPoint(startId + 3);

    outputLines->InsertNextCell(2);
    outputLines->InsertCellPoint(startId + 3);
    outputLines->InsertCellPoint(startId + 0);

    // diagonals (eigenvectors)
    outputLines->InsertNextCell(2);
    outputLines->InsertCellPoint(startId + 0);
    outputLines->InsertCellPoint(startId + 2);

    outputLines->InsertNextCell(2);
    outputLines->InsertCellPoint(startId + 1);
    outputLines->InsertCellPoint(startId + 3);
  }
  
  // colors
  outputColors->InsertNextValue(color);
  outputColors->InsertNextValue(color);
  outputColors->InsertNextValue(color);
  outputColors->InsertNextValue(color);
}

void generateGlyph(double posX, double posY, double posZ,
		   vec3 eigenvalues,
		   vec3 eigenvectors[3],
		   bool allReal,
		   double glyphRadius,
		   bool normalizeGlyph,
		   bool transparentGlyph,
		   vtkPoints *outputPoints,
		   vtkCellArray *outputLines,
		   vtkCellArray *outputPolygons,
		   vtkFloatArray *outputColors,
		   int &numPoints)
{ // glyph for critical points

  // ... call generateLine() and generateQuad() for generating the glyph

#if 0 // code snippet:
  // in the case of complex eigenvalues, 
  // get two vectors spanning the 2D manifold (complex eigenplane)
  vec3 a, b;
  vec3ortho(eigenvectors[1], b);
  vec3cross(b, eigenvectors[1], a);
#endif
}

bool boundaryCell(vtkUnstructuredGrid *grid, int cell)
{
  bool onBoundary = false;
  for (int face=0; face<4; face++) {
    
    // get face points
    vtkIdList *ptIds = vtkIdList::New();
    ptIds->SetNumberOfIds(3);
    int pntCnt = 0;
    for (int p=0; p<4; p++) {
      if (p != face) {
	ptIds->SetId(pntCnt++, grid->GetCell(cell)->GetPointId(p));
      }
    }
    
    // lookup cell that shares the current face
    vtkIdList *cellIds = vtkIdList::New();
    cellIds->SetNumberOfIds(1);
    grid->GetCellNeighbors(cell, ptIds, cellIds);
    
    onBoundary = cellIds->GetNumberOfIds() < 1;
    
    ptIds->Delete();
    cellIds->Delete();
    
    if (onBoundary) {
      // boundary cell
      break;;
    }
  }

  return onBoundary;
}

bool candidateCell(vtkUnstructuredGrid *grid,
		   vtkFloatArray *inVectors,
		   int cell,
		   int zeroEdgesUComp[4],
		   int &zeroEdgeUCnt)
{ // for early rejection of cells

  // computes tetrahedron edges containing a zero of the first vector component,
  // outputting as Idx in this edge list: (0,1) (0,2) (0,3) (1,2) (2,3) (3,0)

  // get node IDs of cell
  vtkIdList *cellNodeIds = grid->GetCell(cell)->GetPointIds();

  // go over vector components
  bool candidate = true;
  for (int comp=0; comp<3; comp++) {
    
    // count tetrahedron edges that contain a zero
    int zeroEdgeCnt = 0;
    
    // go over 3 edges sharing point 0
    float vec0[3];
    inVectors->GetTupleValue(cellNodeIds->GetId(0), vec0);
    for (int node=1; node<4; node++) {
      
      float vecW[3];
      inVectors->GetTupleValue(cellNodeIds->GetId(node), vecW);
      
      if (vec0[comp] * vecW[comp] < 0.0) {
	
	// current edge contains a zero

	if ((comp == 0) && (zeroEdgeCnt < 3)) {
	  zeroEdgesUComp[zeroEdgeCnt] = node - 1;
	  zeroEdgeUCnt = zeroEdgeCnt + 1;
	}

	zeroEdgeCnt++;
      }
    }
    
    // go over 3 edges not sharing point 0
    for (int node=1; node<4; node++) {
      
      int nextNode = node + 1;
      if (nextNode > 3) nextNode = 1;

      float vec1[3], vec2[3];
      inVectors->GetTupleValue(cellNodeIds->GetId(node), vec1);      
      inVectors->GetTupleValue(cellNodeIds->GetId(nextNode), vec2);      
      
      if (vec1[comp] * vec2[comp] < 0.0) {
	
	// current edge contains a zero

	if ((comp == 0) && (zeroEdgeCnt < 3)) {
	  zeroEdgesUComp[zeroEdgeCnt] = node + 2;
	  zeroEdgeUCnt = zeroEdgeCnt + 1;
	}

	zeroEdgeCnt++;
      }
    }
    
    // require at least 3 zero edges per component
    // (early rejection of degenerate cases, TODO: handle these cases)
    if (zeroEdgeCnt < 3) {
      candidate = false;
      break;
    }
  }
  return candidate;
}

bool computeFace(vtkUnstructuredGrid *grid,
		 vtkFloatArray *inVectors,
		 int cell,
		 int zeroEdgesUComp[4],
		 int &zeroEdgeUCnt,
		 float triaCorners[3 * 3],
		 float triaCornersVec[3 * 3])
{ // compute zero level isosurface of first vector component

  // get node IDs of cell
  vtkIdList *cellNodeIds = grid->GetCell(cell)->GetPointIds();

  // compute triangular isosurface of first (u) vector component on tetrahedron
  for (int edge=0; edge<3; edge++) {
    
    // get point indices of current tetrahedron edge
    int pnt1Idx, pnt2Idx;
    if (zeroEdgesUComp[edge] < 3) {
      pnt1Idx = 0;
      pnt2Idx = zeroEdgesUComp[edge] + 1;
    }
    else {
      pnt1Idx = zeroEdgesUComp[edge] - 2;
      pnt2Idx = pnt1Idx + 1;
      if (pnt2Idx > 3) {
	pnt2Idx = 1;
      }
    }

    // get first component at points
    float vec1[3], vec2[3];
    inVectors->GetTupleValue(cellNodeIds->GetId(pnt1Idx), vec1); 
    inVectors->GetTupleValue(cellNodeIds->GetId(pnt2Idx), vec2); 

    // find linear interpolation parameter s
    double s = vec1[0] / (vec1[0] - vec2[0]);

    // get positions of points
    float *positions = 
      ((vtkFloatArray *) grid->GetPoints()->GetData())->GetPointer(0);
    float *pos1 = &positions[cellNodeIds->GetId(pnt1Idx)*3];
    float *pos2 = &positions[cellNodeIds->GetId(pnt2Idx)*3];
    
    // interpolate position of zero point
    triaCorners[edge*3 + 0] = (1.0 - s) * pos1[0] + s * pos2[0];
    triaCorners[edge*3 + 1] = (1.0 - s) * pos1[1] + s * pos2[1];
    triaCorners[edge*3 + 2] = (1.0 - s) * pos1[2] + s * pos2[2];

    // interpolate vector (only second component needed)
    triaCornersVec[edge*3 + 0] = (1.0 - s) * vec1[0] + s * vec2[0];
    triaCornersVec[edge*3 + 1] = (1.0 - s) * vec1[1] + s * vec2[1];
    triaCornersVec[edge*3 + 2] = (1.0 - s) * vec1[2] + s * vec2[2]; 
  }

  // abort if there are no zeros of the second vector component on triangle
  int zeroCnt = 0;
  for (int edge=0; edge<3; edge++) {

    int pnt1Idx = edge;
    int pnt2Idx = edge + 1;
    if (pnt2Idx > 2) pnt2Idx = 0;

    // compare signs
    if (triaCornersVec[pnt1Idx*3+1] * triaCornersVec[pnt2Idx*3+1] < 0.0) {
      zeroCnt++;
    }
  }

  // require 2 edges containing a zero
  // this rejects degenerate cases, TODO: handle these cases
  if (zeroCnt != 2) {
    return false;
  }
  return true;
}

bool computeLine(float triaCorners[3 * 3],
		 float triaCornersVec[3 * 3],
		 float lineEnds[2 * 3],
		 float lineEndsVec[2 * 3])
{ // compute zero level contour of second vector component on triangle

  int zeroCnt = 0;
  for (int edge=0; edge<3; edge++) {

    int pnt1Idx = edge;
    int pnt2Idx = edge + 1;
    if (pnt2Idx > 2) pnt2Idx = 0;

    // compare signs
    if (triaCornersVec[pnt1Idx*3+1] * triaCornersVec[pnt2Idx*3+1] < 0.0) {
      
      // find linear interpolation parameter s
      double s = triaCornersVec[pnt1Idx*3+1] / (triaCornersVec[pnt1Idx*3+1] - triaCornersVec[pnt2Idx*3+1]);
      
      // interpolate position of zero point
      lineEnds[zeroCnt*3 + 0] =
	(1.0 - s) * triaCorners[pnt1Idx*3 + 0] + s * triaCorners[pnt2Idx*3 + 0];
      lineEnds[zeroCnt*3 + 1] =
	(1.0 - s) * triaCorners[pnt1Idx*3 + 1] + s * triaCorners[pnt2Idx*3 + 1];
      lineEnds[zeroCnt*3 + 2] =
	(1.0 - s) * triaCorners[pnt1Idx*3 + 2] + s * triaCorners[pnt2Idx*3 + 2];
      
      // interpolate vector (only third component needed)
      lineEndsVec[zeroCnt*3+0] =
	(1.0 - s) * triaCornersVec[pnt1Idx*3+0] + s * triaCornersVec[pnt2Idx*3+0];
      lineEndsVec[zeroCnt*3+1] =
	(1.0 - s) * triaCornersVec[pnt1Idx*3+1] + s * triaCornersVec[pnt2Idx*3+1];
      lineEndsVec[zeroCnt*3+2] =
	(1.0 - s) * triaCornersVec[pnt1Idx*3+2] + s * triaCornersVec[pnt2Idx*3+2];
      
      zeroCnt++;
    }
  }

  // test if third vector component has different signs on line ends
  if (lineEndsVec[0*3+2] * lineEndsVec[1*3+2] >= 0.0) {
    // same signs -> reject
    return false;
  }
  return true;
}

void computePoint(float lineEnds[2 * 3],
		  float lineEndsVec[2 * 3],
		  float point[3],
		  float pointVec[3])
{ // get point at zero level of third vector component on line

  // find linear interpolation parameter s
  double s = lineEndsVec[0*3+2] / (lineEndsVec[0*3+2] - lineEndsVec[1*3+2]);
  
  // interpolate position of zero point
  point[0] = (1.0 - s) * lineEnds[0*3 + 0] + s * lineEnds[1*3 + 0];
  point[1] = (1.0 - s) * lineEnds[0*3 + 1] + s * lineEnds[1*3 + 1];
  point[2] = (1.0 - s) * lineEnds[0*3 + 2] + s * lineEnds[1*3 + 2];
  
  // interpolate vector
  pointVec[0] = (1.0 - s) * lineEndsVec[0*3 + 0] + s * lineEndsVec[1*3 + 0];
  pointVec[1] = (1.0 - s) * lineEndsVec[0*3 + 1] + s * lineEndsVec[1*3 + 1];
  pointVec[2] = (1.0 - s) * lineEndsVec[0*3 + 2] + s * lineEndsVec[1*3 + 2];
  
#if 1 // debug
  double vecMag =
    sqrt(pointVec[0]*pointVec[0] +
	 pointVec[1]*pointVec[1] +
	 pointVec[2]*pointVec[2]);
  
  printf("found crit pnt at (%g, %g, %g), vecMag=%g\n",
	 point[0], point[1], point[2], vecMag);
#endif
}

bool findCritPoint(vtkUnstructuredGrid *grid,
		   vtkFloatArray *inVectors,
		   int cell,
		   int zeroEdgesUComp[4],
		   int zeroEdgesUCnt,
		   float point[3])
{
  if (zeroEdgesUCnt != 3) {
    // TODO: support quadrangles
    printf("skipping unsupported quadrangle isosurface\n");
    return false;
  }

  // get zero level isosurface of first vector component
  // actually only supporting triangles, TODO: generalize to quadrangle
  float triaCorners[3 * 3];
  float triaCornersVec[3 * 3];
  if (!computeFace(grid, inVectors, cell, zeroEdgesUComp, zeroEdgesUCnt,
		   triaCorners, triaCornersVec)) {
    return false;
  }

  // get zero level line on isosurface of second vector component
  float lineEnds[2 * 3];
  float lineEndsVec[2 * 3];
  if (!computeLine(triaCorners, triaCornersVec, lineEnds, lineEndsVec)) {
    return false;
  }

  // get point at zero level of third vector component on line
  float pointVec[3];
  computePoint(lineEnds, lineEndsVec, point, pointVec);
  
  return true;
}

void computeVelocityGradient(vtkUnstructuredGrid *grid,
			     vtkFloatArray *inVectors,
			     int cell,
			     mat3 gradientOutput)
{ // compute velocity gradient (Jacobian) inside given cell 

  // get node IDs of cell
  vtkIdList *cellNodeIds = grid->GetCell(cell)->GetPointIds();
  
  // get positions and velocities at nodes
  vec3 nodesPos[4];
  vec3 nodesVec[4];
  for (int n=0; n<4; n++) {
    float wVec[3];
    inVectors->GetTupleValue(cellNodeIds->GetId((n+3) % 4), wVec);
    vec3set(nodesVec[n], wVec[0], wVec[1], wVec[2]);

    float *positions = 
      ((vtkFloatArray *) grid->GetPoints()->GetData())->GetPointer(0);
    float *pos = &positions[cellNodeIds->GetId((n+3)%4)*3];
    vec3set(nodesPos[n], pos[0], pos[1], pos[2]);
  }

  // compute inverse grid Jacobian
  mat3 mi;
  {
    mat3 m;
    vec3sub(nodesPos[1], nodesPos[0], m[0]);
    vec3sub(nodesPos[2], nodesPos[0], m[1]);
    vec3sub(nodesPos[3], nodesPos[0], m[2]);
    mat3inv(m, mi);
  }

  // compute weights
  vec3 wt[4];
  {
    mat3 mit;
    mat3trp(mi, mit);

    vec3copy(mit[0], wt[1]);
    vec3copy(mit[1], wt[2]);
    vec3copy(mit[2], wt[3]);

    vec3add(wt[1], wt[2], wt[0]);
    vec3add(wt[0], wt[3], wt[0]);
    vec3scal(wt[0], -1.0, wt[0]);
  }

  // compute gradient
  {
    mat3zero(gradientOutput);
    
    for (int i=0; i<4; i++) {
      mat3 m;
      vec3scal(wt[i], nodesVec[i][0], m[0]);
      vec3scal(wt[i], nodesVec[i][1], m[1]);
      vec3scal(wt[i], nodesVec[i][2], m[2]);
      mat3add(gradientOutput, m, gradientOutput);
    }
  }
}

bool analyzeCritPoint(vtkUnstructuredGrid *grid,
		      vtkFloatArray *inVectors,
		      int cell,
		      float point[3],
		      vec3 eigenvalues,
		      vec3 eigenvectors[3])
{ // returns true if all eigenvalues are real

  // compute velocity gradient (Jacobian)
  mat3 grad;
  computeVelocityGradient(grid, inVectors, cell, grad);

  // compute eigenvectors and eigenvalues
  bool allReal;
  {
    allReal = (mat3eigenvalues(grad, eigenvalues) == 3);

#if 1 // debug
    if (allReal) {
      printf("  eigenvalues: %g, %g, %g\n",
	     eigenvalues[0], eigenvalues[1], eigenvalues[2]);
    }
    else {
      printf("  eigenvalues: %g, %g +-i %g\n",
	     eigenvalues[0], eigenvalues[1], eigenvalues[2]);
    }
#endif

    // compute real eigenvector
    mat3realEigenvector(grad, eigenvalues[0], eigenvectors[0]);
      
    if (allReal) {
      // real eigenvalues -> compute eigenvectors of 2D manifold
      mat3realEigenvector(grad, eigenvalues[1], eigenvectors[1]);
      mat3realEigenvector(grad, eigenvalues[2], eigenvectors[2]);
    }
    else {
      // complex eigenvalues -> compute normal of complex eigenplane
      mat3complexEigenplane(grad, eigenvalues[1], eigenvalues[2], eigenvectors[1]);
    }
  }

  return allReal;
}

void vtkFlowTopo::computeIt(vtkUnstructuredGrid *grid,
			    vtkFloatArray *inVectors,
			    vtkPolyData *polyDat)
{ 
  int geomType;
  vtkPolyData *vtkPolyDat = polyDat;
  vtkPoints *outputPoints;
  vtkCellArray *outputLines;
  vtkCellArray *outputPolygons;
  vtkDataArray *outputColors;
  int numPoints = 0;
  //int numLines = 0;
  int numCritPoints = 0;

  outputPoints = vtkPoints::New();
  outputLines = vtkCellArray::New();
  outputPolygons = vtkCellArray::New();
  outputColors = vtkFloatArray::New();

  // go over unstructured cells
  for (int c=0; c<grid->GetNumberOfCells(); c++) {

    if (grid->GetCellType(c) != VTK_TETRA) {
      printf("skipping cell %d because cell type not supported\n", c);
    }

    // skip tetrahedra at boundary
    if (skipBoundaryCells) {
      if (boundaryCell(grid, c)) continue;
    }

    // early rejection of cells without critical points
    // also evaluates which edges contain a zero of the first vector component
    int zeroEdgesUComp[4];
    int zeroEdgesUCnt;
    if (!candidateCell(grid, inVectors, c, zeroEdgesUComp, zeroEdgesUCnt)) {
      // skip cell
      continue;
    }

    // find critical point in cell
    float point[3];
    if (!findCritPoint(grid, inVectors, c, zeroEdgesUComp, zeroEdgesUCnt, point)) {
      // no point found -> skip
      continue;
    }

    numCritPoints++;

    // analyze critical point
    vec3 eigenvalues, eigenvectors[3];
    bool allReal = analyzeCritPoint(grid, inVectors, c, point, eigenvalues, eigenvectors);

    // generate glyph
    generateGlyph(point[0], point[1], point[2],
		  eigenvalues, eigenvectors, allReal,
		  GlyphRadius, NormalizeGlyph, TransparentGlyph,
		  outputPoints, outputLines, outputPolygons,
		  (vtkFloatArray *) outputColors, numPoints);
  }

  printf("%d critical points\n", numCritPoints);

  // set colors
  outputColors->SetName("Colors");
  int idx = vtkPolyDat->GetPointData()->AddArray(outputColors);
  vtkPolyDat->GetPointData()->SetActiveAttribute(idx, vtkDataSetAttributes::SCALARS);

  // set geometry
  vtkPolyDat->SetPoints(outputPoints);
  vtkPolyDat->SetLines(outputLines);
  vtkPolyDat->SetPolys(outputPolygons);
 
  // reference is counted, we can delete
  outputPoints->Delete();
  outputLines->Delete();
  outputPolygons->Delete();
  outputColors->Delete();
}

int vtkFlowTopo::RequestData(
  vtkInformation *vtkNotUsed(request),
  vtkInformationVector **inputVector,
  vtkInformationVector *outputVector)
{
  // get the info objects
  vtkInformation *inInfo = inputVector[0]->GetInformationObject(0);
  vtkInformation *outInfo = outputVector->GetInformationObject(0);

  // get the input and output
  vtkUnstructuredGrid *input =
    vtkUnstructuredGrid::SafeDownCast(inInfo->Get(vtkDataObject::DATA_OBJECT()));
  vtkPolyData *output =
    vtkPolyData::SafeDownCast(outInfo->Get(vtkDataObject::DATA_OBJECT()));

  vtkIdType numCells = input->GetNumberOfCells();

  // TODO: support also Double
  vtkFloatArray *inVectors = (vtkFloatArray *) this->GetInputArrayToProcess(0, inputVector);

  // Initialize
  vtkDebugMacro(<<"Computing flow topology (critical points)");

  // Check input
  if (numCells < 1) { 
    printf("no cells\n");
    vtkErrorMacro("No cells");
    return 1;
  }

  if (!inVectors) {
    printf("no vector data\n");
    vtkDebugMacro(<<"No vector data");
    return 1;
  }
  
  // compute
  computeIt(input, inVectors, output);
 
  return 1;
}

int vtkFlowTopo::FillInputPortInformation(int, vtkInformation *info)
{
  info->Set(vtkAlgorithm::INPUT_REQUIRED_DATA_TYPE(), "vtkDataSet");
  return 1;
}

void vtkFlowTopo::PrintSelf(ostream& os, vtkIndent indent)
{
  this->Superclass::PrintSelf(os,indent);
}
